/* -*- c-basic-offset: 3; mode: c++ -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 *
 * 
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

/**
 * Compute global discrepancy error values for individual discrepancy error values.
 */

#include <iomanip>
#include <math.h>

#include <pandore.h>
using namespace pandore;

inline void
computeDetectionAccuracyEvaluationResult( DiscrepancyIndicator &indicator, EvaluationResult & result, int criteria ) {
   const int nbOfErrors = 2;
   for (int i = 0; i < nbOfErrors; i++) {
      if (indicator.errors[i].normalization == 0) {
	 result.errors[criteria * 2 + i] = 0;
      } else {
	 result.errors[criteria * 2 + i] = indicator.errors[i].value / indicator.errors[i].normalization;
      }
   }
}

inline void
computeFragmentationConsistencyEvaluationResult( DiscrepancyIndicator &indicator, EvaluationResult & result, int criteria ) {
   const int nbOfErrors = 2;   
   for (int i = 0; i < nbOfErrors; i++) {
      if (indicator.errors[i].normalization == 0) {
	 result.errors[criteria * 2 + i] = 0;
      } else {
	 Float H = indicator.errors[i].value / indicator.errors[i].normalization;
	 result.errors[criteria * 2 + i] =  H / (H + 1.0f);
      }   
   }
}

inline void
computeBoundaryPrecisionEvaluationResult( DiscrepancyIndicator &indicator, EvaluationResult & result, int criteria ) {
   const int nbOfErrors = 2;
   for (int i = 0; i < nbOfErrors; i++) {
      if (indicator.errors[i].normalization == 0) {
	 result.errors[criteria * 2 + i] = 0;
      } else {
	 result.errors[criteria * 2 + i] = indicator.errors[i].value / indicator.errors[i].normalization;
      }
   }
}

inline void
computeShapeFidelityEvaluationResult( DiscrepancyIndicator &indicator, EvaluationResult & result, int criteria ) {
   const int nbOfErrors = 2;
   for (int i = 0; i < nbOfErrors; i++) {
      if (indicator.errors[i].normalization == 0) {
	 result.errors[criteria * 2 + i] = 0;
      } else {
	 float x = log10(indicator.errors[i].value / indicator.errors[i].normalization  + 1);
	 result.errors[criteria * 2 + i] =  x / (1.0F + x);
      }
   }
}

inline void
computeTopologyPreservationResult( DiscrepancyIndicator &indicator, EvaluationResult & result, int criteria ) {
   const int nbOfErrors = 2;
   for (int i = 0; i < nbOfErrors; i++) {
      if (indicator.errors[i].normalization == 0) {
	 result.errors[criteria * 2 + i] = 0;
      } else {
	 result.errors[criteria * 2 + i] = indicator.errors[i].value / indicator.errors[i].normalization;
      }
   }
}

inline Errc
ComputeErrorValues( const  DiscrepancyResult * indicatorsForAllSegmentationResults, int vectorSize,
		    EvaluationResult & evaluationResult) {
   std::string type1, type2;

   const std::string nameError = "Error";
   const std::string nameNormalization = "Normalization";

   int criteria;
   
   // 1. detection accuracy
   criteria = 0;
   {
      DiscrepancyIndicator indicator;
      
      for (int e = 0; e < 2; e++) {
	 for (int i = 0; i < vectorSize; ++i) {
	    indicator.errors[e].value += indicatorsForAllSegmentationResults[i].indicators[criteria].errors[e].value;
	    indicator.errors[e].normalization += indicatorsForAllSegmentationResults[i].indicators[criteria].errors[e].normalization;
	 }
      }
      computeDetectionAccuracyEvaluationResult(indicator, evaluationResult, criteria);
      
   }

   // 2. fragmentation consistency
   criteria++;
   {
      DiscrepancyIndicator indicator;
      for (int e = 0; e < 2; e++) {
	 for (int i = 0; i < vectorSize; ++i) {
	    indicator.errors[e].value += indicatorsForAllSegmentationResults[i].indicators[criteria].errors[e].value;
	    indicator.errors[e].normalization += indicatorsForAllSegmentationResults[i].indicators[criteria].errors[e].normalization;
	 }
      }
      computeFragmentationConsistencyEvaluationResult(indicator, evaluationResult, criteria);

   }

   // 3. boundary precision
   criteria++;
   {
      DiscrepancyIndicator indicator;
      for (int e = 0; e < 2; e++) {
	 for (int i = 0; i < vectorSize; ++i) {
	    indicator.errors[e].value += indicatorsForAllSegmentationResults[i].indicators[criteria].errors[e].value;
	    indicator.errors[e].normalization += indicatorsForAllSegmentationResults[i].indicators[criteria].errors[e].normalization;
	 }
      }
      computeBoundaryPrecisionEvaluationResult(indicator, evaluationResult, criteria);
   }

   // 4. shape fidelity
   criteria++;
   {
      DiscrepancyIndicator indicator;
      for (int e = 0; e < 2; e++) {
	 for (int i = 0; i < vectorSize; ++i) {
	    indicator.errors[e].value += indicatorsForAllSegmentationResults[i].indicators[criteria].errors[e].value;
	    indicator.errors[e].normalization += indicatorsForAllSegmentationResults[i].indicators[criteria].errors[e].normalization;
	 }
      }
      computeShapeFidelityEvaluationResult(indicator, evaluationResult, criteria);
   }

   // 5. topology preservation
   criteria++;
   {
      DiscrepancyIndicator indicator;
      for (int e = 0; e < 2; e++) {
	 for (int i = 0; i < vectorSize; ++i) {
	    indicator.errors[e].value += indicatorsForAllSegmentationResults[i].indicators[criteria].errors[e].value;
	    indicator.errors[e].normalization += indicatorsForAllSegmentationResults[i].indicators[criteria].errors[e].normalization;
	 }
      }
      computeTopologyPreservationResult(indicator, evaluationResult, criteria);
   }

   return SUCCESS;
}
