/* -*- c-basic-offset: 3; mode: c++ -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 *
 * 
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

#include <iomanip>
#include <algorithm>
#include "utils.h"

#include "./regiongraphmatching.h"
#include "./detectionaccuracymeasurement.h"
#include "./fragmentationconsistencymeasurement.h"
#include "./boundaryprecisionmeasurement.h"
#include "./shapefidelitymeasurement.h"
#include "./topologypreservationmeasurement.h"
#include "segmentation/plabeling.cpp"


inline int
medianIndex( Float * errors, int size ) {
   Float *tmp = new Float[size];
   for (int i = 0; i < size; i++ ) {
      tmp[i] = errors[i];
   }
   // tri
   float* first(&tmp[0]);
   float* last(first + size);
   std::sort(first, last);

   Float median = tmp[(size-1)/2];
   int index = 0;

   for (int i = 0; i < size; i++ ) {
      if (median == errors[i]) {
	 index = i;
	 break;
      }
   }
   delete[] tmp;
   return index;
}

inline void
computeMedianErrorsAgainstReferences( DiscrepancyResult *groundtruthIndicators, int size, DiscrepancyResult &imageIndicator ) {
   Float *errors1 = new Float[size];
   Float *errors2 = new Float[size];

   for (int i = 0; i < 5; i++) {
      for (int gt = 0; gt < size; gt++) {
	 Float error1;
	 if (groundtruthIndicators[gt].indicators[i].errors[0].normalization == 0) {
	    error1 = 0;
	 } else {
	    error1 = groundtruthIndicators[gt].indicators[i].errors[0].value / groundtruthIndicators[gt].indicators[i].errors[0].normalization;
	 }
	 errors1[gt] = error1;
	 Float error2; 
	 if (groundtruthIndicators[gt].indicators[i].errors[1].normalization == 0) {
	    error2 = 0;
	 } else {
	    error2 = groundtruthIndicators[gt].indicators[i].errors[1].value / groundtruthIndicators[gt].indicators[i].errors[1].normalization;
	 }
	 errors2[gt] = error2;
      }
      imageIndicator.indicators[i].errors[0] = groundtruthIndicators[medianIndex(errors1, size)].indicators[i].errors[0];
      imageIndicator.indicators[i].errors[1] = groundtruthIndicators[medianIndex(errors2, size)].indicators[i].errors[1];
   }
   delete[] errors1;
   delete[] errors2;
}

inline bool
computePerformanceValues( int matching_algorithm_id,
			  float matching_threshold,
			  const std::string & segmentation_image_name,
			  const std::string & ground_truth_image_name,
			  DiscrepancyResult *evaluation) {
   Graph2d grs;
   Reg2d segmentation_image;
   Reg2d ground_truth_image;
   
   if (!segmentation_image.LoadFile(segmentation_image_name.c_str())) {
      std::cerr << "Error: file " << segmentation_image_name.c_str() << "is not a regular Pandore region map" << std::endl;
	return false;
   }
   if (!ground_truth_image.LoadFile(ground_truth_image_name.c_str())) {
	std::cerr << "Error: file " << ground_truth_image_name.c_str() << "is not a regular Pandore region map" << std::endl;
	return false;
   }
   
   try {
      RegionGraphMatching(segmentation_image, ground_truth_image, grs, matching_algorithm_id, matching_threshold);
   } catch (PandoreException& caught)  {
      std::cerr << "Error evaluatealgorithm: " << caught.what() << std::endl;
      return false;
   }
   
   Graph2d *gri;
   
   gri = (Graph2d*)grs.Clone();
   evaluation->indicators[0] = DetectionAccuracyMeasurement(segmentation_image, ground_truth_image, *gri);
   gri->Delete();
   
   gri = (Graph2d*)grs.Clone();
   evaluation->indicators[1] = FragmentationConsistencyMeasurement(segmentation_image, ground_truth_image, *gri);
   gri->Delete();
   
   gri = (Graph2d*)grs.Clone();
   evaluation->indicators[2] = BoundaryPrecisionMeasurement(segmentation_image, ground_truth_image, *gri);
   gri->Delete();
   
   gri = (Graph2d*)grs.Clone();
   evaluation->indicators[3] = ShapeFidelityMeasurement(segmentation_image, ground_truth_image, *gri);
   gri->Delete();
   
   gri = (Graph2d*)grs.Clone();
   evaluation->indicators[4] = TopologyPreservationMeasurement(segmentation_image, ground_truth_image, *gri, 8);
   gri->Delete();
   
   LOG(INFO) << "  " << std::fixed << std::setprecision(2);
   for (int i = 0; i < 5; ++i) {
      LOG(INFO) << "(" << evaluation->indicators[i].errors[0].value << " / "<< evaluation->indicators[i].errors[0].normalization
   		<< ", " << evaluation->indicators[i].errors[1].value << " / "<< evaluation->indicators[i].errors[1].normalization << ") ";
   }
   LOG(INFO) << std::endl;
   
   return true;
}

inline Errc
EvaluateAlgorithm( int matching_algorithm_id, float matching_threshold,
		   std::string segmentation_result_path, std::string ground_truth_path,
		   DiscrepancyResult * & indicatorsForAllSegmentationResults ) {
   if (!endsWith(segmentation_result_path, PATH_SEPARATOR)) {
      segmentation_result_path += PATH_SEPARATOR;
   }
   std::vector<std::string> files = std::vector<std::string>();
   if (!getAllFilesWithExtension(segmentation_result_path, ".pan", files)) {
//      std::cerr << "[ERROR: " << errno << " ] No segmentation result in" << segmentation_result_path << "." << std::endl;
      exit(1);
   }
   if (files.size() == 0) {
      std::cerr << "No segmentation result in directory: " << segmentation_result_path << std::endl;
      exit(2);
   }

   if (!endsWith(ground_truth_path, PATH_SEPARATOR)) {
      ground_truth_path += PATH_SEPARATOR;
   }
   std::vector<std::string> directories = std::vector<std::string>();
   getSubdirectories(ground_truth_path, directories);
   if (directories.size() == 0) {
      std::cerr << "Error: No expert subdirectory found in directory: " << ground_truth_path << std::endl;
      return FAILURE;
   }

   std::sort(files.begin(), files.end());
   std::sort(directories.begin(), directories.end() );

   indicatorsForAllSegmentationResults = new DiscrepancyResult[files.size()];
   DiscrepancyResult *indicatorsForOneSegmentationResult = new DiscrepancyResult[directories.size()];
   int nbImages = 0;
   for (unsigned int image = 0; image < files.size(); image++) {
      std::string directory = dirname(files[image]);
      std::string subdirectory = getlastdirectory(directory.c_str(), segmentation_result_path.c_str());
      std::string imagename = basename(files[image]); 
      std::string segmentation_image = files[image];

      // Error measures against all EXPERTS
      int nbExperts = 0;
      // for (unsigned int expert = 0; expert < directories.size(); expert++) {
      for (unsigned int expert = 0; expert < directories.size(); expert++) {
	 std::string expert_directory = getlastdirectory(directories[expert].c_str(), ground_truth_path.c_str());
	 std::string ground_truth_image = ground_truth_path + expert_directory + PATH_SEPARATOR + subdirectory + PATH_SEPARATOR + imagename;
	 if (isFile(segmentation_image) && isFile(ground_truth_image)) {
	    LOG(INFO) << std::endl << "Segmentation result: " << segmentation_image << std::endl;
	    LOG(INFO) << "Ground truth image:  " << ground_truth_image << std::endl;
	    if (computePerformanceValues(matching_algorithm_id, matching_threshold,
					 segmentation_image, ground_truth_image,
					 &indicatorsForOneSegmentationResult[nbExperts])) {
	       nbExperts++;
	       // use the final result as the median errors against all experts.
	    }
	 }
      }
      if (nbExperts > 0) {
	 computeMedianErrorsAgainstReferences(indicatorsForOneSegmentationResult, nbExperts, indicatorsForAllSegmentationResults[nbImages++]);
      }
   }

   return nbImages;
}
