/* -*- c-basic-offset: 3; mode: c++ -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 *
 *
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

/**
 * The consistency measure consists of fragmentation and inclusion indices.
 * It is based on the conditional entropy measures.
 *
 * H(R|S) = Sum Sum (|Ri * Sj| / n) . log2(|Sj| / |Ri * Sj|)
 * where * is the intesection operator, and n is the total overlapping area.
 *
 * Reference : Lukashevich 2008, Gong 2011.
 */

#include "evaluation.h"
#include <cmath>

#ifdef WIN32
// Calculates log2 of number.  
inline float
log2( float n ) {
    return (float)(log(n) / log(2.0));
}
#endif

inline DiscrepancyError
computeUnderSegmentationError( const MatchingGraph & graph, int numberOfSegments ) {
   // 1. Compute actual overlapped segment areas (Sj inter {Ri})
   for (int s = 1; s < numberOfSegments; s++) {
      long total = 0;      
      if (graph.isInverseSegmentMatched(s)) { // Overlapping
	 for (GEdge *ptr = graph.getInverseMatchedRegionsList(s); ptr != NULL; ptr = ptr->Next()) {
	    total += (long)ptr->weight;
#ifndef NDEBUG
	    printf(" s=%d : r=%d area=%f total = %ld\n", s, ptr->Node() - numberOfSegments, ptr->weight, graph.getSegmentArea(s));
#endif
	 }
	 graph.setSegmentArea(s, total);
      }
   }

   int segmentCount = 0;
   float Hrs = 0;
   float totalArea = 0;
   float totalOverlap = 0;
   // 2. Compute conditional entropy
   for (int s = 1; s < numberOfSegments; s++) {
      float entropy = 0;
      if (graph.getSegmentArea(s) > 0 && graph.isInverseSegmentMatched(s)) {
	 segmentCount++;
	 for (GEdge *ptr = graph.getInverseMatchedRegionsList(s); ptr != NULL; ptr = ptr->Next()) {
	    float prs = (float)(ptr->weight / graph.getSegmentArea(s));
	    entropy += prs * log2(prs);
	    totalOverlap += (float)ptr->weight;
	 }
	 Hrs += graph.getSegmentArea(s) * entropy;
	 totalArea += (float)graph.getSegmentArea(s);
#ifndef NDEBUG
	 printf("  segment area: %ld  entropy: %f Hrs=%f\n", graph.getSegmentArea(s), entropy, graph.getSegmentArea(s) * entropy);
#endif
      }
   }
   
   if (segmentCount > 0 && Hrs != 0) {
      Hrs /= (float)- totalOverlap;
   } else {
      Hrs = 0;
   }
   
#ifndef NDEBUG
   float inclusion = Hrs / (1.0 + Hrs);
   printf("* FRAGMENTATION CONSISTENCY\n");
   printf(" error 3/  Under-segmentation..........: %.3f (%.1f region(s) per correct segment in average)\n", fabs(inclusion), exp(Hrs * log(2.0)));
#endif
   return DiscrepancyError(Hrs, 1);
}

inline DiscrepancyError
computeOverSegmentationError( const MatchingGraph & graph, int numberOfRegions ) {
   // float fragmentation = 0;
   // Compute actual overlapped region areas (do not consider missed pixels) 
   for (int r = 1; r < numberOfRegions; r++) {
      long total = 0;
      if (graph.isInverseRegionMatched(r)) { // Overlapping.
	 for (GEdge *ptr = graph.getInverseMatchedSegmentsList(r); ptr != NULL; ptr = ptr->Next()) {
	    total += (Long)ptr->weight;
#ifndef NDEBUG
	    printf(" r=%d : s=%d area=%f total = %ld\n", r, ptr->Node(), ptr->weight, graph.getRegionArea(r));
#endif
	 }
	 graph.setRegionArea(r, total);
      }
   }
   
   // Compute overlapped region areas
   int regionCount = 0;
   float Hsr = 0;
   float totalArea = 0;
   float totalOverlap = 0;
   
   for (int r = 1; r < numberOfRegions; r++) {
      float entropy = 0;
      if (graph.getRegionArea(r) > 0 && graph.isInverseRegionMatched(r)) {
	 regionCount++;
	 for (GEdge *ptr = graph.getInverseMatchedSegmentsList(r); ptr != NULL; ptr = ptr->Next()) {
	    float prs = (float)(ptr->weight / graph.getRegionArea(r));
	    entropy += prs * log2(prs);
	    totalOverlap += (float)ptr->weight;
	 }
	 Hsr += graph.getRegionArea(r) * entropy;
	 totalArea += (float)graph.getRegionArea(r);
#ifndef NDEBUG
	 printf("  region area: %ld  entropy: %f Hsr=%f\n", graph.getRegionArea(r), entropy, graph.getRegionArea(r) * entropy);
#endif
      }
   }
   
   if (regionCount > 0 && Hsr != 0) {
      Hsr /= (float) -totalOverlap;
   } else {
      Hsr = 0;
   }
#ifndef NDEBUG
   float fragmentation = Hsr / (1.0 + Hsr);
   printf(" error 4/  Over-segmentation...........: %.3f (%.1f segment(s) per detected region in average)\n", fabs(fragmentation), exp(Hsr * log(2.0)));
#endif
   
   return DiscrepancyError(Hsr, 1);
}

inline DiscrepancyIndicator
FragmentationConsistencyMeasurement( const Reg2d &segmentation, const Reg2d &reference,
				     Graph2d &grs ) {
   if (segmentation.Size() != reference.Size()) {
      throw PandoreException("Error "__FILE__": Incompatible size between the two region maps.");
   }
   
   const int numberOfSegments = segmentation.Labels() + 1;
   const int numberOfRegions = reference.Labels() + 1;
   
   DiscrepancyIndicator indicator;
   MatchingGraph graph(numberOfSegments, numberOfRegions);
   graph.setMatchingGraph(&grs);
   
   indicator.errors[0] = computeUnderSegmentationError(graph, numberOfSegments);   
   indicator.errors[1] = computeOverSegmentationError(graph, numberOfRegions);
   
   return indicator;
}
