/* -*- c-basic-offset: 3; mode: c++ -*-
 *
 * Copyright (c), GREYC.
 * All rights reserved
 *
 * You may use this file under the terms of the BSD license as follows:
 *
 * "Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 *   * Neither the name of the GREYC, nor the name of its
 *     contributors may be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
 *
 * 
 * For more information, refer to:
 * https://clouard.users.greyc.fr/Pandore
 */

/**
 * References: Average Hausdorff based on Dubuisson et Jain.
 */

#include "evaluation.h"
#include <cmath>

inline void
getRiBoundary( const MatchingGraph &graph,
	       const Reg2d &reference,
	       std::vector<Point2d> *regionBoundaries ) {
   Point2d p;
   
   for (p.y = 0; p.y < reference.Height(); p.y++) {
      for (p.x = 0; p.x < reference.Width(); p.x++) {
	 int r = reference[p];
	 if (r > 0 && graph.isRegionMatched(r)) {
	    for (int v = 0; v < 4; v++) {
	       bool isContourPoint =  !reference.Hold(p + v4[v]) || reference[p + v4[v]] != reference[p];
	       if (isContourPoint) {
		  regionBoundaries[r].push_back(p);
		  break;
	       }
	    }
	 }
      }
   }
}

inline void
getRiInterSjBoundary( const MatchingGraph &graph,
		      const Reg2d &reference,
		      const Reg2d &segmentation,
		      std::vector<Point2d> *regionInterSegmentsBoundaries ) {
   Point2d p;
   
   for (p.y = 0; p.y < reference.Height(); p.y++) {
      for (p.x = 0; p.x < reference.Width(); p.x++) {
	 int r = reference[p];
	 if (r > 0 && graph.isRegionMatched(r)) {
	    int s = segmentation[p];
	    if (graph.isRegionMatchedBy(r, s)) {
	       for (int v = 0; v < 4; v++) {
		  bool isContourPoint = ((!reference.Hold(p + v4[v]))
				    || ((reference[p + v4[v]] != (Ulong)r)
					|| ((segmentation[p + v4[v]] != (Ulong)s)
					    && !graph.isRegionMatchedBy(r, segmentation[p + v4[v]]))));
		  if (isContourPoint) {
		     regionInterSegmentsBoundaries[r].push_back(p);
		     break;
		  }
	       }
	    }
	 }
      }
   }
}

inline DiscrepancyError
computeModifiedHausdorffDistanceBetweenRiAndRiInterSj( const MatchingGraph &graph,
						       const std::vector<Point2d> *regionBoundaries,
						       const std::vector<Point2d> *regionInterSegmentBoundaries,
						       const Reg2d &segmentation, 
						       int numberOfRegions ) {
   const float maxDistance = (float)segmentation.VectorSize() * segmentation.VectorSize();
   float MHD = 0;
   int numberOfTruePositiveRegions = 0;

   for (int r = 1; r < numberOfRegions; r++) {
      if (graph.isRegionMatched(r)) {
	 int pixelCount11 = 0;
	 float totalDistance11 = 0;
    
	 for (std::vector<Point2d>::const_iterator itr = regionBoundaries[r].begin();
	      itr != regionBoundaries[r].end();
	      itr++) {
	    Point2d p((*itr).y, (*itr).x);
	    const int s = segmentation[p];
	    if (!graph.isRegionMatchedBy(r, s)) {
	       float minDistance = maxDistance;
	       for (std::vector<Point2d>::const_iterator its = regionInterSegmentBoundaries[r].begin();
		    its != regionInterSegmentBoundaries[r].end();
		    its++) {
		  int x = ABS(p.x - (*its).x);
		  int y = ABS(p.y - (*its).y);
		  float distance = ((float)x * x + (float)y * y);
		  if (distance < minDistance) {
		     minDistance = distance;
		  }
		  if (minDistance == 0.0f) {
		     break;
		  }
	       }
	       if (minDistance > 0) {
		  pixelCount11++;
		  totalDistance11 += sqrt(minDistance);
	       }
	    }
	 }
	 float MHD1 = (pixelCount11 == 0)? 0 : totalDistance11 / pixelCount11;

	 int pixelCount12 = 0;
	 float totalDistance12 = 0;
	 for (std::vector<Point2d>::const_iterator its = regionInterSegmentBoundaries[r].begin();
	      its != regionInterSegmentBoundaries[r].end();
	      its++) {
	    Point2d p((*its).y, (*its).x);
	    float minDistance = maxDistance;
	    for (std::vector<Point2d>::const_iterator itr = regionBoundaries[r].begin();
		 itr != regionBoundaries[r].end();
		 itr++) {
	       int x = ABS(p.x - (*itr).x);
	       int y = ABS(p.y - (*itr).y);
	       float distance = ((float)x * x + (float)y * y);
	       if (distance < minDistance) {
		  minDistance = distance;
	       }
	       if (minDistance == 0.0f) {
		  break;
	       }
	    }
	    if (minDistance > 0) {
	       pixelCount12++;
	       totalDistance12 += sqrt(minDistance);
	    }
	 }
	 float MHD2 = (pixelCount12 == 0)? 0.0f : totalDistance12 / pixelCount12;

	 float mhd = std::max(MHD1, MHD2);
	 if (mhd > 0) {
	    numberOfTruePositiveRegions++;
	    MHD += mhd;
	 }
      }
   }

#ifndef NDEBUG
   printf("\n* SHAPE FIDELITY\n");
   float v = 0;
   if (numberOfTruePositiveRegions> 0) {
      v = (MHD / numberOfTruePositiveRegions);
   }
   float dmean1 = (Float)(log10(v + 1) / (1.0 + log10(v + 1)));
   printf(" 7/  Shape omission error......: %.3f (%.2f pixels away from the segment boundary in average)\n", v, dmean1);
#endif

   return DiscrepancyError(MHD, (float)numberOfTruePositiveRegions);
}

inline void
getSiBoundary( const MatchingGraph &graph,
	       const Reg2d &segmentation,
	       std::vector<Point2d> *segmentBoundaries ) {
   Point2d p;
   
   for (p.y = 0; p.y < segmentation.Height(); p.y++) {
      for (p.x = 0; p.x < segmentation.Width(); p.x++) {
	 int s = segmentation[p];
	 if (s > 0 &&  graph.isSegmentMatched(s)) {
	    for (int v = 0; v < 4; v++) {
	       bool isContourPoint = !segmentation.Hold(p + v4[v]) || segmentation[p + v4[v]] != segmentation[p];
	       if (isContourPoint) {
		  segmentBoundaries[s].push_back(p);
		  break;
	       }
	    }
	 }
      }
   }
}

inline void
getSiInterRjBoundary( const MatchingGraph &graph,
		      const Reg2d &segmentation,
		      const Reg2d &reference,
		      std::vector<Point2d> *segmentInterRegionBoundaries ) {
   Point2d p;
   for (p.y = 0; p.y < segmentation.Height(); p.y++) {
      for (p.x = 0; p.x < segmentation.Width(); p.x++) {
	 int s = segmentation[p];
	 if (s>0 && graph.isSegmentMatched(s)) {
	    int r = reference[p];
	    if (graph.isSegmentMatchedBy(s, r)) {
	       for (int v = 0; v < 4; v++) {
		  bool isContourPoint = ((!segmentation.Hold(p + v4[v]))
				    || ((segmentation[p + v4[v]] != (Ulong)s)
					|| ((reference[p + v4[v]] != (Ulong)r)
					    && !graph.isSegmentMatchedBy(s, reference[p + v4[v]]))));
		  if (isContourPoint) {
		     segmentInterRegionBoundaries[s].push_back(p);
		     break;
		  }
	       }
	    }
	 }
      }
   }
}

inline DiscrepancyError
computeModifiedHausdorffDistanceBetweenSiAndSiInterRj( const MatchingGraph &graph,
						       const std::vector<Point2d> *segmentBoundaries,
						       const std::vector<Point2d> *segmentInterRegionBoundaries,
						       const Reg2d &reference, 
						       int numberOfSegments ) {
   const float maxDistance = (float)reference.VectorSize()*reference.VectorSize();
   float MHD = 0;
   int numberOfTruePositiveSegments = 0;
   
   for (int s = 1; s < numberOfSegments; s++) {
      if (graph.isSegmentMatched(s)) {
	 int pixelCount21 = 0;
	 float totalDistance21 = 0;
	 for (std::vector<Point2d>::const_iterator its = segmentBoundaries[s].begin();
	      its != segmentBoundaries[s].end();
	      its++) {
	    Point2d p((*its).y, (*its).x);
	    const int r = reference[p];
	    if (!graph.isSegmentMatchedBy(s, r)) {
	       float minDistance = maxDistance;
	       for (std::vector<Point2d>::const_iterator itr = segmentInterRegionBoundaries[s].begin();
		    itr != segmentInterRegionBoundaries[s].end();
		    itr++) {
		  int x = ABS(p.x - (*itr).x);
		  int y = ABS(p.y - (*itr).y);
		  float distance = ((float)x * x + (float)y * y);
		  if (distance < minDistance) {
		     minDistance = distance;
		  }
		  if (minDistance == 0.0f) {
		     break;
		  }
	       }
	       if (minDistance > 0) {
		  pixelCount21++;
		  totalDistance21 += sqrt(minDistance);
	       }
	    }
	 }
	 float MHD1 = (pixelCount21 == 0)? 0 : totalDistance21 / pixelCount21;
    
	 int pixelCount22 = 0;
	 float totalDistance22 = 0;
	 for (std::vector<Point2d>::const_iterator itr = segmentInterRegionBoundaries[s].begin();
	      itr != segmentInterRegionBoundaries[s].end();
	      itr++) {
	    Point2d p((*itr).y, (*itr).x);
	    float minDistance = maxDistance;
	    for (std::vector<Point2d>::const_iterator its = segmentBoundaries[s].begin();
		 its != segmentBoundaries[s].end();
		 its++) {
	       int x = ABS(p.x - (*its).x);
	       int y = ABS(p.y - (*its).y);
	       float distance = ((float)x * x + (float)y * y);
	       if (distance < minDistance) {
		  minDistance = distance;
	       }
	       if (minDistance == 0.0f) {
		  break;
	       }
	    }
	    if (minDistance > 0) {
	       pixelCount22++;
	       totalDistance22 += sqrt(minDistance);
	    }
	 }
	 float MHD2 =  (pixelCount22 == 0)? 0.0f : totalDistance22 / pixelCount22;

	 float mhd = std::max(MHD1, MHD2);
	 if (mhd > 0) {
	    numberOfTruePositiveSegments++;
	    MHD += mhd;
	 }
      }
   }

#ifndef NDEBUG
   float v = 0;
   if (numberOfTruePositiveSegments > 0) {
      v = (MHD / numberOfTruePositiveSegments);
   }
   float dmean1 = (Float)(log10(v + 1)/(1.0 + log10(v + 1)));
   printf(" 8/  Shape commission error....: %.3f (%.2f pixels away from the region boundary in average)\n", v, dmean1);
#endif
   return DiscrepancyError(MHD, (float)numberOfTruePositiveSegments);
}

inline DiscrepancyIndicator
ShapeFidelityMeasurement( const Reg2d &segmentation, const Reg2d &reference,
			  Graph2d &grs ) {
   if (segmentation.Size() != reference.Size()) {
      throw PandoreException("Error "__FILE__": Incompatible size between the two region maps.");
   }

   const int numberOfSegments = segmentation.Labels() + 1;
   const int numberOfRegions = reference.Labels() + 1;

   DiscrepancyIndicator indicator;
   MatchingGraph graph(numberOfSegments, numberOfRegions);
   
   graph.setMatchingGraph(&grs);
   
   std::vector<Point2d> *regionBoundaries = new std::vector<Point2d>[numberOfRegions];
   getRiBoundary(graph, reference, regionBoundaries);
   
   std::vector<Point2d> *regionInterSegmentBoundaries = new std::vector<Point2d>[numberOfRegions];
   getRiInterSjBoundary(graph, reference, segmentation, regionInterSegmentBoundaries);
   
   indicator.errors[0] = computeModifiedHausdorffDistanceBetweenRiAndRiInterSj(graph,
									    regionBoundaries,
									    regionInterSegmentBoundaries,
									    segmentation,
									    numberOfRegions);
   
   delete[] regionBoundaries;
   delete[] regionInterSegmentBoundaries;
   
   std::vector<Point2d> *segmentBoundaries = new std::vector<Point2d>[numberOfSegments];
   getSiBoundary(graph, segmentation, segmentBoundaries);
   
   std::vector<Point2d> *segmentInterRegionBoundaries = new std::vector<Point2d>[numberOfSegments];
   getSiInterRjBoundary(graph, segmentation, reference, segmentInterRegionBoundaries);
   
   indicator.errors[1] = computeModifiedHausdorffDistanceBetweenSiAndSiInterRj(graph,
									    segmentBoundaries,
									    segmentInterRegionBoundaries,
									    reference, 
									    numberOfSegments);
   
   delete[] segmentBoundaries;
   delete[] segmentInterRegionBoundaries;

   return indicator;
}
